# Flow

General purpose graph package for Go

## Initial Idea

`possibilities`

* Machine learning
* Graphical CI pipeline

## Flow package

Using flow package without UI

```go
package main

import (
	"flow"
	"flow/registry"
	"log"
	"strings"
)

func main() {

	r := registry.New()
	r.Add(strings.Split, strings.Join)

	f := flow.New()
	f.SetRegistry(r)

	op := f.Op("Join",
		f.Op("Split", "hello world", " "),
		",",
	)

	res, err := op.Process()
	if err != nil {
		log.Fatal(err)
	}
	log.Println("res:", res)

}
```

## How it works

A flow works by requesting the previous nodes the results of its operation,
so the starting node will be the node we want the result of, unattached nodes wont be executed

### Registering a function

All nodes are Go functions to create a node src simply add a function like:

```go
package main

import (
	"flow/flowserver"
	"flow/registry"
	"net/http"
)

func main() {

	r := registry.New()
	r.Add("hello", func() string {
		return "hello world"
	})

	http.ListenAndServe(":5000", flowserver.New(r, "storename"))

}
```

Resulting in

![fig1](img/s1.jpg)

### Go funcs

Since every node is a function we can register functions from any Go
package

```go
// Adding functions from a go package
r.Add(strings.Split, strings.Join)
```

![fig2](img/gofuncs.jpg)

### Describing funcs

We can describe the inserted functions either by using `r.Add` methods or
grouping functions with the helper `registry.Describer`

```go
// utility to apply functions to several entries
registry.Describer(
	r.Add(strings.Split).Inputs("str", "sep").Output("slice"),
	r.Add(strings.Join).Inputs("slice", "sep").Output("string"),
).Tags("strings").Extra("style", registry.M{"color": "#a77"})
```

![fig3](img/describing.jpg)

### Results

![fig4](img/runningsplit.jpg)
![fig5](img/runningjoin.jpg)

### Special Operations

#### f.In

Uses an argument passed through the process function

```go
op := f.In(0)
res, err := op.Process("arg")
// res will be "arg"
```

#### f.Var

Variable that stores data in the flow (for future serialization) it can be
restored

```go
f.Var("myvar", "initial value or operation")
```

#### f.SetVar

Set var creates an operation that sets a variable to the result of the operation
passed as the argument

```go
f.SetVar("myvar", operation)
```

### UI Special nodes

#### Input

Same as f.In where a property must be set to indicate which argument this
node refers to

#### Output

Flow-UI only output is a special node that indicates the overall flow output,
the UI contains a RUN button that will be visible if there is an output node and
the flow will be executed from the output node.

There can be only one output node

#### Var

Same as f.Var passing the variable name as a property of the node

#### SetVar

Same as f.SetVar passing the variable name as a property of the node

#### Portal From

UI Only: Portals are helper nodes that allows to connect areas of UI without
crossing any links, right click in a node and choose create portal

---

> WIP

## TODO

### UX/UI

* UX/UI: Create Undo behaviour
* UX/UI: Special visualisers/nodes to display information (images, datatables, graphs)
* UX/UI: Ability to group nodes into a single box exposing inputs and outputs
* UX/UI: Implement touch
* UX/UI: Drop link in node to link to the next compatible input
* ~~UX/UI: Portals to clean graph crossing~~ (testing)
* UX/UI: `Shift` key to merge group selections

### Packages

* FlowPkg: Create training mechanism
* FlowPkg: Machine learning examples
* FlowPkg: Generate go code to maximize call performance

### Backend

* Backend: Export/Deploy a model to a training cluster
* Backend: Expose the model possibly via an Rest API for easy function access

### Other

* Collaboration: Better concurrent editing/message passing
* Flow: Proper documentation

## Ideas

### Matching types

in UI we can create a state such as draggingType{in:'io.Writer'} which basically
each node can figure a match for each socket

in Server side we can do an analysis on each type to see if it can be
implemented by another type and send this as part of the registry with all
descriptions, this way we can highlight sockets that can implement the dragging
type

### Generating code

generate code based on a flow should be simple as we have the function signatures

### Portals

> Testing phase

Named portal would connect a node to another without any link, this way we can
have clean links without crossovers

### Recursive registering

Since an operation can be run with op.Process(inputs...) we can simple do
registry.Add(op.Process) to register a flow as an operation, UI doesn't retain
the registry/flow yet
